import numpy as np
import tensorflow as tf
import tensorflow.keras.layers as tfkl
from tensorflow.keras.utils import custom_object_scope

from dynastes import object_scope
from dynastes.blocks.attention_blocks import SelfAttentionBlock1D
from dynastes.blocks.transformer_blocks import PointWiseFeedForwardBlock, EncoderBlock, EncoderBlockStack


def _test_grads(testCase: tf.test.TestCase, func, input):
    _, grads = tf.test.compute_gradient(func, input)
    for grad in grads:
        testCase.assertNotAllClose(grad, np.zeros_like(grad))
        testCase.assertAllInRange(grad, -400., 400)


to_tensor = tf.convert_to_tensor
normal = np.random.normal


class EncoderBlockTest(tf.test.TestCase):
    def test_simple(self):
        with custom_object_scope(object_scope):
            d_model = 16
            num_heads = 4
            dff = 32
            length = 4096
            mask_len = 32
            sablock = SelfAttentionBlock1D(d_model, d_model, num_heads=num_heads,
                                           attention_type='PseudoBlockSparseAttention1D',
                                           block_length=1024,
                                           multiquery_attention=True,)
            norm = tfkl.LayerNormalization(epsilon=1e-6)
            df_net = PointWiseFeedForwardBlock(dff=dff, d_model=d_model)
            enc_block = EncoderBlock(sa_layer=sablock, norm0=norm, ffn=df_net, norm1=norm)
            stack = EncoderBlockStack([enc_block] * 3)

            test_input = tf.convert_to_tensor(normal(size=(1, length, d_model)).astype(np.float32))
            mask = to_tensor(([True] * (length - mask_len)) + ([False] * (mask_len)))
            mask = tf.expand_dims(mask, axis=0)
            out = stack(test_input, training=None, mask=mask)

            comp_out_shape = stack.compute_output_shape(test_input.shape)

            self.assertEqual(out.shape, comp_out_shape)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .spectral import SpectralNormalization
from .util import WscaleNormalizer, get, serialize, deserialize, object_scope

# Cleanup symbols to avoid polluting namespace.
del absolute_import
del division
del print_function



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from . import backend
from .nn import array_ops
from .nn import math_ops
from .nn import nn_ops

# Cleanup symbols to avoid polluting namespace.
del absolute_import
del division
del print_function



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .attention_layers import LocalizedAttentionLayer1D
from .attention_layers import LocalizedAttentionLayer2D
from .base_layers import ActivatedKernelBiasBaseLayer, _WscaleInitializer, DynastesDense
from .conditioning_layers import FeaturewiseLinearModulation
from .convolutional_layers import DynastesConv1D
from .convolutional_layers import DynastesConv1DTranspose
from .convolutional_layers import DynastesConv2D
from .convolutional_layers import DynastesConv2DTranspose
from .convolutional_layers import DynastesConv3D
from .convolutional_layers import DynastesDepthwiseConv1D
from .convolutional_layers import DynastesDepthwiseConv2D
from .normalization_layers import AdaptiveGroupNormalization
from .normalization_layers import AdaptiveInstanceNormalization
from .normalization_layers import AdaptiveLayerInstanceNormalization
from .normalization_layers import AdaptiveLayerNormalization
from .normalization_layers import AdaptiveMultiNormalization
from .normalization_layers import MultiNormalization
from .normalization_layers import PoolNormalization2D
from .random_layers import StatelessRandomNormalLike
from .t2t_attention_layers import AddTimingSignalLayer1D
from .t2t_attention_layers import Attention1D
from .t2t_attention_layers import Attention2D
from .t2t_attention_layers import LshGatingLayer
from .t2t_attention_layers import PseudoBlockSparseAttention1D
from .time_delay_layers import DepthGroupwiseTimeDelayLayer1D
from .time_delay_layers import DepthGroupwiseTimeDelayLayerFake2D
from .time_delay_layers import TimeDelayLayer1D
from .time_delay_layers import TimeDelayLayerFake2D

object_scope = {
    '_WscaleInitializer': _WscaleInitializer,
    'DynastesDense': DynastesDense,
    'LocalizedAttentionLayer1D': LocalizedAttentionLayer1D,
    'LocalizedAttentionLayer2D': LocalizedAttentionLayer2D,
    'ActivatedKernelBiasBaseLayer': ActivatedKernelBiasBaseLayer,
    'DepthGroupwiseTimeDelayLayer1D': DepthGroupwiseTimeDelayLayer1D,
    'DepthGroupwiseTimeDelayLayerFake2D': DepthGroupwiseTimeDelayLayerFake2D,
    'TimeDelayLayer1D': TimeDelayLayer1D,
    'TimeDelayLayerFake2D': TimeDelayLayerFake2D,
    'AddTimingSignalLayer1D': AddTimingSignalLayer1D,
    'LshGatingLayer': LshGatingLayer,
    'Attention1D': Attention1D,
    'Attention2D': Attention2D,
    'PseudoBlockSparseAttention1D': PseudoBlockSparseAttention1D,
    'StatelessRandomNormalLike': StatelessRandomNormalLike,
    'PoolNormalization2D': PoolNormalization2D,
    'MultiNormalization': MultiNormalization,
    'AdaptiveMultiNormalization': AdaptiveMultiNormalization,
    'AdaptiveLayerInstanceNormalization': AdaptiveLayerInstanceNormalization,
    'AdaptiveInstanceNormalization': AdaptiveInstanceNormalization,
    'AdaptiveGroupNormalization': AdaptiveGroupNormalization,
    'AdaptiveLayerNormalization': AdaptiveLayerNormalization,
    'FeaturewiseLinearModulation': FeaturewiseLinearModulation,
    'DynastesConv1D': DynastesConv1D,
    'DynastesConv2D': DynastesConv2D,
    'DynastesConv3D': DynastesConv3D,
    'DynastesConv1DTranspose': DynastesConv1DTranspose,
    'DynastesConv2DTranspose': DynastesConv2DTranspose,
    'DynastesDepthwiseConv1D': DynastesDepthwiseConv1D,
    'DynastesDepthwiseConv2D': DynastesDepthwiseConv2D,
}

# Cleanup symbols to avoid polluting namespace.
del absolute_import
del division
del print_function



import tensorflow as tf
from tensorflow.keras.utils import custom_object_scope

from dynastes import object_scope
from dynastes.layers.convolutional_layers import DynastesConv1DTranspose, DynastesDepthwiseConv1D, \
    DynastesConv2DTranspose, DynastesConv2D, DynastesDepthwiseConv2D, DynastesConv1D
from dynastes.util.test_utils import layer_test

import numpy as np
to_tensor = tf.convert_to_tensor
normal = np.random.normal

class DynastesConv1DTest(tf.test.TestCase):

    def test_masking(self):
        with custom_object_scope(object_scope):

            layer = DynastesConv1D(32, kernel_size=3, strides=2, padding='same')

            ts = to_tensor(normal(size=(8, 16, 32))
                          .astype(np.float32))

            mask_len = 16 // 6
            mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
            mask = tf.expand_dims(mask, axis=0)
            mask = tf.tile(mask, [8, 1])
            layer(ts, mask=mask)
            layer.compute_mask(ts, mask=mask)

            @tf.function
            def graph_test_fn(x, mask):
                layer(x, mask=mask)
                layer.compute_mask(x, mask=mask)

            graph_test_fn(x=ts, mask=mask)


    def test_simple(self):
        with custom_object_scope(object_scope):
            layer_test(
                DynastesConv1D, kwargs={'filters': 3, 'kernel_size': 3}, input_shape=(5, 32, 3))

    def test_specnorm(self):
        with custom_object_scope(object_scope):
            layer_test(
                DynastesConv1D, kwargs={'filters': 3, 'kernel_size': 3,
                                        'kernel_normalizer': 'spectral',
                                        'use_wscale': True,
                                        'kernel_regularizer': 'orthogonal'}, input_shape=(5, 32, 3))


class DynastesConv2DTest(tf.test.TestCase):
    def test_simple(self):
        with custom_object_scope(object_scope):
            layer_test(
                DynastesConv2D, kwargs={'filters': 3, 'kernel_size': (3, 3)}, input_shape=(4, 16, 16, 3))

    def test_masking(self):
        with custom_object_scope(object_scope):

            layer = DynastesConv2D(32, kernel_size=3, strides=2, padding='same')

            ts = to_tensor(normal(size=(8, 16, 16, 32))
                          .astype(np.float32))

            mask_len = 16 // 8
            mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
            mask = tf.expand_dims(mask, axis=0)
            mask = tf.tile(mask, [8, 1])
            mask = tf.expand_dims(mask, axis=1)
            mask = tf.tile(mask, [1, 16, 1])
            layer(ts, mask=mask)
            layer.compute_mask(ts, mask=mask)

            @tf.function
            def graph_test_fn(x, mask):
                layer(x, mask=mask)
                layer.compute_mask(x, mask=mask)

            graph_test_fn(x=ts, mask=mask)

    def test_specnorm(self):
        with custom_object_scope(object_scope):
            layer_test(
                DynastesConv2D, kwargs={'filters': 7, 'kernel_size': (3, 3),
                                        'kernel_normalizer': 'spectral',
                                        'kernel_regularizer': 'orthogonal'}, input_shape=(4, 16, 16, 5))


class DynastesConv1DTransposeTest(tf.test.TestCase):
    def test_simple(self):
        with custom_object_scope(object_scope):
            layer_test(
                DynastesConv1DTranspose,
                kwargs={'filters': 3, 'kernel_size': 3, 'strides': 2, 'padding': 'same'},
                input_shape=(None, 16, 3),
                expected_output_shape=(None, 32, 3)
            )

    def test_masking(self):
        with custom_object_scope(object_scope):

            layer = DynastesConv1DTranspose(32, kernel_size=3, strides=2, padding='same')

            ts = to_tensor(normal(size=(8, 16, 32))
                          .astype(np.float32))

            mask_len = 16 // 6
            mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
            mask = tf.expand_dims(mask, axis=0)
            mask = tf.tile(mask, [8, 1])
            layer(ts, mask=mask)
            layer.compute_mask(ts, mask=mask)

            @tf.function
            def graph_test_fn(x, mask):
                layer(x, mask=mask)
                layer.compute_mask(x, mask=mask)

            graph_test_fn(x=ts, mask=mask)

    def test_specnorm(self):
        with custom_object_scope(object_scope):
            layer_test(
                DynastesConv1DTranspose,
                kwargs={'filters': 7, 'kernel_size': 3, 'strides': 2, 'padding': 'same',
                        'kernel_normalizer': 'spectral',
                        'kernel_regularizer': 'orthogonal'},
                input_shape=(None, 16, 5),
                expected_output_shape=(None, 32, 7)
            )

class DynastesConv2DTransposeTest(tf.test.TestCase):
    def test_simple(self):
        with custom_object_scope(object_scope):
            layer_test(
                DynastesConv2DTranspose,
                kwargs={'filters': 3, 'kernel_size': (3, 3), 'strides': (2, 2), 'padding': 'same'},
                input_shape=(None, 16, 16, 3),
                expected_output_shape=(None, 32, 32, 3)
            )

    def test_masking(self):
        with custom_object_scope(object_scope):

            layer = DynastesConv2DTranspose(32, kernel_size=3, strides=2, padding='same')

            ts = to_tensor(normal(size=(8, 16, 16, 32))
                          .astype(np.float32))

            mask_len = 16 // 8
            mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
            mask = tf.expand_dims(mask, axis=0)
            mask = tf.tile(mask, [8, 1])
            mask = tf.expand_dims(mask, axis=1)
            mask = tf.tile(mask, [1, 16, 1])
            layer(ts, mask=mask)
            layer.compute_mask(ts, mask=mask)

            @tf.function
            def graph_test_fn(x, mask):
                layer(x, mask=mask)
                layer.compute_mask(x, mask=mask)

            graph_test_fn(x=ts, mask=mask)

    def test_specnorm(self):
        with custom_object_scope(object_scope):
            layer_test(
                DynastesConv2DTranspose,
                kwargs={'filters': 7, 'kernel_size': (3, 3), 'strides': (2, 2), 'padding': 'same',
                        'kernel_normalizer': 'spectral',
                        'kernel_regularizer': 'orthogonal'},
                input_shape=(None, 16, 16, 5),
                expected_output_shape=(None, 32, 32, 7)
            )


class DynastesDepthwiseConv1DTest(tf.test.TestCase):
    def test_simple(self):
        with custom_object_scope(object_scope):
            layer_test(
                DynastesDepthwiseConv1D, kwargs={'kernel_size': 3}, input_shape=(5, 32, 3))

    def test_masking(self):
        with custom_object_scope(object_scope):

            layer = DynastesDepthwiseConv1D(kernel_size=3, strides=2, padding='same')

            ts = to_tensor(normal(size=(8, 16, 32))
                          .astype(np.float32))

            mask_len = 16 // 6
            mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
            mask = tf.expand_dims(mask, axis=0)
            mask = tf.tile(mask, [8, 1])
            layer(ts, mask=mask)
            layer.compute_mask(ts, mask=mask)

            @tf.function
            def graph_test_fn(x, mask):
                layer(x, mask=mask)
                layer.compute_mask(x, mask=mask)

            graph_test_fn(x=ts, mask=mask)

    def test_specnorm(self):
        with custom_object_scope(object_scope):
            layer_test(
                DynastesDepthwiseConv1D, kwargs={'kernel_size': 3,
                                                 'kernel_normalizer': 'spectral',
                                                 'kernel_regularizer': 'orthogonal'}, input_shape=(5, 32, 3))


class DynastesDepthwiseConv2DTest(tf.test.TestCase):
    def test_simple(self):
        with custom_object_scope(object_scope):
            layer_test(
                DynastesDepthwiseConv2D, kwargs={'kernel_size': (3, 3)}, input_shape=(4, 16, 16, 3))

    def test_masking(self):
        with custom_object_scope(object_scope):

            layer = DynastesDepthwiseConv2D(kernel_size=3, strides=2, padding='same')

            ts = to_tensor(normal(size=(8, 16, 16, 32))
                          .astype(np.float32))

            mask_len = 16 // 6
            mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
            mask = tf.expand_dims(mask, axis=0)
            mask = tf.tile(mask, [8, 1])
            mask = tf.expand_dims(mask, axis=1)
            mask = tf.tile(mask, [1, 16, 1])
            layer(ts, mask=mask)
            layer.compute_mask(ts, mask=mask)

            @tf.function
            def graph_test_fn(x, mask):
                layer(x, mask=mask)
                layer.compute_mask(x, mask=mask)

            graph_test_fn(x=ts, mask=mask)


    def test_specnorm(self):
        with custom_object_scope(object_scope):
            layer_test(
                DynastesDepthwiseConv2D, kwargs={'kernel_size': (3, 3),
                                                 'kernel_normalizer': 'spectral'}, input_shape=(4, 16, 16, 5))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .attention_blocks import SelfAttentionBlock1D

object_scope = {
    'SelfAttentionBlock1D': SelfAttentionBlock1D,
}

# Cleanup symbols to avoid polluting namespace.
del absolute_import
del division
del print_function
